#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

import ssmash

# -- General configuration ---------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ["sphinx.ext.autodoc", "sphinx.ext.viewcode"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"ssmash"
copyright = u"2019, Gary Donovan"
author = u"Gary Donovan"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = ssmash.__version__
# The full version, including alpha/beta/rc tags.
release = ssmash.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "ssmashdoc"

# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "ssmash", u"ssmash Documentation", [author], 1)]



"""Functions to create Custom Resources in Cloud Formation"""


# NB: Don't put imports here for Lambda function handlers. They should be self-contained


# TODO tests


def restart_ecs_service_resource_handler(event, context):
    """Lambda handler function to restart an ECS service, as a CloudFormation resource.

    This is intended to be used in an inline deployment. The physical ID is
    the ID of the ECS service deployment
    """
    # TODO consider bringing the `cfnresponse` module in locally as a pseudo-dependency, so that we can reference it.
    # The `cfnresponse` module is injected at runtime if you are executing a Custom Resource handler
    # noinspection PyUnresolvedReferences
    import cfnresponse

    import logging

    logging.basicConfig(level=logging.DEBUG)
    LOGGER = logging.getLogger("restarter")  # TODO
    try:
        import boto3

        # TODO is there a way to make this idempotent on event["RequestId"]? OR just hope it only gets called once...

        ecs = boto3.client("ecs")
        response = {}
        physical_id = event.get("PhysicalResourceId")

        if event["RequestType"] in ["Create", "Update"]:
            # Restart the ECS service
            #   FIXME this breaks if multiple restarts get scheduled
            # FIXME Not sure how best to deal with timeouts... don't forget context.get_remaining_time_in_millis()
            properties = event["ResourceProperties"]
            update_result = ecs.update_service(
                cluster=properties["ClusterArn"],
                service=properties["ServiceArn"],
                forceNewDeployment=True,
            )

            # Get the deployment ID, which is first in the list. TODO pluck out newest using deployment["createdAt"]
            deployment = update_result["service"]["deployments"][0]
            # assert deployment["status"] == "PRIMARY" #TODO real error or don't bother
            physical_id = deployment["id"]

            # Poll until the ECS service reaches a steady state
            ecs.get_waiter("services_stable").wait(
                cluster=properties["ClusterArn"],
                services=[properties["ServiceArn"]],
                # TODO use WaiterConfig to timeout inside out expected timeout
            )

            # FIXME check result from ecs waiter
        elif event["RequestType"] == "Delete":
            # Doesn't make any sense to delete an ECS deployment - just return
            LOGGER.info(
                "Ignoring request to delete resource for deployment %s", physical_id
            )
        else:
            raise ValueError("Unknown CloudFormation request type")
        cfnresponse.send(event, context, cfnresponse.SUCCESS, response, physical_id)
        # TODO notify (optional) CFN waiter
    except Exception as ex:
        LOGGER.exception("argh!")  # TODO
        # FIXME the cfnresponse module doesn't let us set the `Reason` in the response
        # TODO set PhysicalResourceId in response if known. at least dont let it use the default log stream name (!)
        cfnresponse.send(event, context, cfnresponse.FAILED, {})



"""Tests for the command line interface."""
import re
from contextlib import contextmanager
from datetime import datetime
from datetime import timezone
from unittest.mock import ANY
from unittest.mock import patch

import pytest
import semver
import yaml
from click.testing import CliRunner
from flyingcircus.intrinsic_function import ImportValue
from freezegun import freeze_time

from ssmash import cli
from ssmash.invalidation import create_ecs_service_invalidation_stack

SIMPLE_INPUT = """foo: bar"""
SIMPLE_OUTPUT_LINE = "Name: /foo"


class TestBasicCLI:
    def test_should_display_help(self):
        runner = CliRunner()
        help_result = runner.invoke(cli.run_ssmash, ["--help"])

        assert help_result.exit_code == 0
        assert re.search(r"--help +Show this message and exit", help_result.output)
        assert re.search(
            r"Convert.*YAML.*application\s+configuration.*CloudFormation.*SSM\s+parameter",
            help_result.output,
            re.DOTALL | re.I,
        )

    @pytest.mark.parametrize(
        ("command", "description"),
        [("invalidate-ecs", r"Invalidate.*cache.*ECS\s+Service")],
    )
    def test_sub_commands_should_display_help(self, command, description):
        runner = CliRunner()
        help_result = runner.invoke(cli.run_ssmash, [command, "--help"])

        assert help_result.exit_code == 0
        assert re.search(r"--help +Show this message and exit", help_result.output)
        assert re.search(description, help_result.output, re.DOTALL | re.I)

    def test_should_exit_cleanly_with_empty_input(self):
        runner = CliRunner()
        result = runner.invoke(cli.run_ssmash)
        assert result.exit_code == 0
        assert not result.stderr_bytes


class TestCloudFormationMetadata:
    def test_stack_should_have_description(self):
        # Setup
        description = "Some lengthy text"

        # Exercise
        runner = CliRunner()
        result = runner.invoke(cli.run_ssmash, args=["--description", description])

        # Verify
        cfn = yaml.safe_load(result.stdout)
        assert cfn["Description"] == description

    def test_output_should_contain_version(self):
        from ssmash import __version__ as package_version

        # Exercise
        runner = CliRunner()
        result = runner.invoke(cli.run_ssmash)

        # Verify
        cfn = yaml.safe_load(result.stdout)
        ssmash_version = cfn["Metadata"]["ssmash"]["version"]
        assert ssmash_version == package_version
        semver.parse(ssmash_version)

    def test_output_should_contain_timestamp(self):
        expected_time = datetime(2019, 5, 22, 1, 2, 3, tzinfo=timezone.utc)

        # Exercise
        runner = CliRunner()

        with freeze_time(expected_time):
            result = runner.invoke(cli.run_ssmash)

        # Verify
        cfn = yaml.safe_load(result.stdout)
        timestamp = cfn["Metadata"]["ssmash"]["generated_timestamp"]
        assert timestamp == "2019-05-22T01:02:03+00:00"


class TestCloudFormationIsProduced:
    def test_should_convert_simple_input_with_default_pipes(self):
        # Exercise
        runner = CliRunner()
        result = runner.invoke(cli.run_ssmash, input=SIMPLE_INPUT)

        # Verify
        assert result.exit_code == 0
        assert not result.stderr_bytes

        assert SIMPLE_OUTPUT_LINE in result.stdout

    def test_should_convert_simple_input_with_files(self):
        input_filename = "input.yaml"
        output_filename = "output.yaml"

        runner = CliRunner()
        with runner.isolated_filesystem():
            with open(input_filename, "w") as f:
                f.write(SIMPLE_INPUT)

            # Exercise
            result = runner.invoke(
                cli.run_ssmash, args=["-i", input_filename, "-o", output_filename]
            )

            # Verify
            assert result.exit_code == 0
            assert not result.stdout
            assert not result.stderr_bytes

            with open(output_filename, "r") as f:
                actual_output = f.read()
            assert SIMPLE_OUTPUT_LINE in actual_output


class TestEcsServiceInvalidation:
    @staticmethod
    @contextmanager
    def patch_create_invalidation_stack():
        # Note that we patch the object imported into the `cli` module, not
        # the original function definition in the `invalidation` module
        with patch(
            "ssmash.cli.create_ecs_service_invalidation_stack",
            wraps=create_ecs_service_invalidation_stack,
        ) as mocked:
            yield mocked

    def run_script_with_invalidation_params(
        self, cluster=None, service=None, role=None, extra_args=None
    ):
        """Execute script with simple input, and ECS service invalidation."""
        args = ["invalidate-ecs"]

        if cluster:
            args.append("--cluster-name")
            args.append(cluster)
        if service:
            args.append("--service-name")
            args.append(service)
        if role:
            args.append("--role-name")
            args.append(role)

        if extra_args:
            args.extend(extra_args)

        runner = CliRunner()
        result = runner.invoke(
            cli.run_ssmash, input=SIMPLE_INPUT, args=args, catch_exceptions=False
        )
        return result

    def test_should_call_invalidation_helper_and_have_parameters_in_output(self):
        # Setup
        cluster = "arn:cluster"
        service = "arn:service"
        role = "arn:role"

        # Exercise
        with self.patch_create_invalidation_stack() as invalidation_mock:
            result = self.run_script_with_invalidation_params(cluster, service, role)

        # Verify
        assert result.exit_code == 0

        invalidation_mock.assert_called_with(
            cluster=cluster, service=service, dependencies=ANY, restart_role=role
        )

        assert cluster in result.stdout
        assert service in result.stdout
        assert role in result.stdout

    def test_should_dereference_cloudformation_imports(self):
        # Setup
        cluster_export = "some-cluster-export"
        service_export = "some-service-export"
        role_export = "some-role-export"

        # Exercise
        with self.patch_create_invalidation_stack() as invalidation_mock:
            result = self.run_script_with_invalidation_params(
                extra_args=[
                    "--cluster-import",
                    cluster_export,
                    "--service-import",
                    service_export,
                    "--role-import",
                    role_export,
                ]
            )

        # Verify
        assert result.exit_code == 0

        invalidation_mock.assert_called_with(
            cluster=ImportValue(cluster_export),
            service=ImportValue(service_export),
            dependencies=ANY,
            restart_role=ImportValue(role_export),
        )

    @pytest.mark.parametrize(
        ("cluster", "service", "role"),
        [
            (None, "arn:service", "arn:role"),
            (None, None, "arn:role"),
            (None, "arn:service", None),
            ("arn:cluster", None, "arn:role"),
            ("arn:cluster", None, None),
            ("arn:cluster", "arn:service", None),
        ],
    )
    def test_should_error_if_not_all_parameters_specified(self, cluster, service, role):
        # Exercise
        with self.patch_create_invalidation_stack() as invalidation_mock:
            result = self.run_script_with_invalidation_params(cluster, service, role)

        # Verify
        assert result.exit_code != 0
        invalidation_mock.assert_not_called()

    @pytest.mark.parametrize("param_name", ["cluster", "service", "role"])
    def test_should_error_if_both_name_and_import_specified(self, param_name):
        # Setup
        cluster = "arn:cluster"
        service = "arn:service"
        role = "arn:role"

        # Exercise
        with self.patch_create_invalidation_stack() as invalidation_mock:
            result = self.run_script_with_invalidation_params(
                cluster,
                service,
                role,
                extra_args=[f"--{param_name}-import", "some-import-name"],
            )

        # Verify
        assert result.exit_code != 0
        invalidation_mock.assert_not_called()



# -*- coding: utf-8 -*-

"""Convert a plain YAML file with application configuration into a CloudFormation template with SSM parameters."""

import sys
from datetime import datetime
from datetime import timezone
from functools import partial
from functools import wraps
from typing import Callable
from typing import Optional
from typing import Union

import click
import yaml
from flyingcircus.core import Stack
from flyingcircus.intrinsic_function import ImportValue
from flyingcircus.service.ssm import SSMParameter

from ssmash.converter import convert_hierarchy_to_ssm
from ssmash.invalidation import create_ecs_service_invalidation_stack

# TODO move helper functions to another module
# TODO tests for helper functions


#: Prefix for specifying a CloudFormation import as a CLI parameter
CFN_IMPORT_PREFIX = "!ImportValue:"


@click.group("ssmash", chain=True, invoke_without_command=True, help=__doc__)
@click.option(
    "-i",
    "--input",
    "--input-file",
    "input_file",
    type=click.File("r"),
    default="-",
    help="Where to read the application configuration YAML file",
)
@click.option(
    "-o",
    "--output",
    "--output-file",
    "output_file",
    type=click.File("w"),
    default="-",
    help="Where to write the CloudFormation template file",
)
@click.option(
    "--description",
    type=str,
    default="Application configuration",
    help="The description for the CloudFormation stack.",
)
def run_ssmash(input_file, output_file, description: str):
    pass


@run_ssmash.resultcallback()
def process_pipeline(processors, input_file, output_file, description: str):
    # Create basic processor inputs
    appconfig = _load_appconfig_from_yaml(input_file)
    stack = _initialise_stack(description)

    # Augment processing functions with default loader and writer
    processors = (
        [_create_ssm_parameters]
        + processors
        + [partial(_write_cfn_template, output_file)]
    )

    # Apply all chained commands
    for processor in processors:
        processor(appconfig, stack)


def appconfig_processor(func: Callable) -> Callable:
    """Decorator to convert a Click command into a custom processor for application configuration."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        def processor(appconfig: dict, stack: Stack):
            return func(appconfig, stack, *args, **kwargs)

        return processor

    return wrapper


@run_ssmash.command(
    "invalidate-ecs",
    options_metavar="(--cluster-name|--cluster-import) CLUSTER "
    "(--service-name|--service-import) SERVICE "
    "(--role-name|--role-import) ROLE ",
)
@click.option(
    "--cluster-name",
    type=str,
    default=None,
    help="The cluster that contains the ECS Service to invalidate (as a name or ARN).",
    metavar="ARN",
)
@click.option(
    "--cluster-import",
    type=str,
    default=None,
    help="Alternatively, specify the cluster as a CloudFormation import.",
    metavar="EXPORT_NAME",
)
@click.option(
    "--service-name",
    type=str,
    default=None,
    help="The ECS Service that depends on this configuration (as a name or ARN).",
    metavar="ARN",
)
@click.option(
    "--service-import",
    type=str,
    default=None,
    help=("Alternatively, specify the ECS Service as a CloudFormation export."),
    metavar="EXPORT_NAME",
)
@click.option(
    "--role-name",
    type=str,
    default=None,
    help="The IAM role to use for invalidating this service (as an ARN).",
    metavar="ARN",
)
@click.option(
    "--role-import",
    type=str,
    default=None,
    help=("Alternatively, specify the IAM role as a CloudFormation export."),
    metavar="EXPORT_NAME",
)
@appconfig_processor
def invalidate_ecs_service(
    appconfig,
    stack,
    cluster_name,
    cluster_import,
    service_name,
    service_import,
    role_name,
    role_import,
):
    """Invalidate the cache in an ECS Service that uses these parameters,
    by restarting the service.
    """
    # Unpack the resource references
    cluster = _get_cfn_resource_from_options("cluster", cluster_name, cluster_import)
    service = _get_cfn_resource_from_options("service", service_name, service_import)
    role = _get_cfn_resource_from_options("role", role_name, role_import)

    # Use a custom Lambda to restart the service iff it's dependent resources
    # have changed
    stack.merge_stack(
        create_ecs_service_invalidation_stack(
            cluster=cluster,
            service=service,
            dependencies=[
                r for r in stack.Resources.values() if isinstance(r, SSMParameter)
            ],
            restart_role=role,
        ).with_prefixed_names("InvalidateEcs")
    )


def _create_ssm_parameters(appconfig: dict, stack: Stack):
    """Create SSM parameters for every item in the application configuration"""
    stack.merge_stack(
        convert_hierarchy_to_ssm(appconfig).with_prefixed_names("SSMParam")
    )


def _get_cfn_resource_from_options(
    option_name: str, arn: Optional[str], export_name: Optional[str]
) -> Union[str, ImportValue]:
    """Get a CloudFormation resource from one of several ways to specify it.

    Parameters:
        arn: The physical name or ARN of the underlying resources
        export_name: The name of a CloudFormation Export that can be
            dereferenced to access the resource.
        option_name: The name of the option, used in CLI error messages.
    """
    if arn and export_name:
        raise click.UsageError(
            f"The {option_name} may not be specified using both a name/ARN and a CloudFormation Export."
        )

    if export_name:
        result = ImportValue(export_name)
    else:
        result = arn

    if not result:
        raise click.UsageError(
            f"The {option_name} must be specified using either a name/ARN, or a CloudFormation Export."
        )
    return result


def _initialise_stack(description: str) -> Stack:
    """Create a basic Flying Circus stack, customised for ssmash"""
    stack = Stack(Description=description)

    from ssmash import __version__

    stack.Metadata["ssmash"] = {
        "generated_timestamp": datetime.now(tz=timezone.utc).isoformat(),
        "version": __version__,
    }
    return stack


def _load_appconfig_from_yaml(input) -> dict:
    """Load a YAML description of the application configuration"""
    appconfig = yaml.safe_load(input)

    # Note that PyYAML returns None for an empty file, rather than an empty
    # dictionary
    if appconfig is None:
        appconfig = {}

    return appconfig


def _write_cfn_template(output, appconfig: dict, stack: Stack):
    """Write the CloudFormation template"""
    output.write(stack.export("yaml"))


if __name__ == "__main__":
    sys.exit(run_ssmash())



import re

from flyingcircus.core import AWSObject
from flyingcircus.intrinsic_function import GetAtt
from flyingcircus.intrinsic_function import Ref
from flyingcircus.service.lambda_ import Function
from flyingcircus.service.ssm import SSMParameter
from flyingcircus.service.ssm import SSMParameterProperties

from ssmash.invalidation import create_ecs_service_invalidation_stack


class TestEcsServiceInvalidation:
    def test_creates_stack_with_supplied_parameters(self):
        # Setup
        ssm_parameter = SSMParameter(
            Properties=SSMParameterProperties(
                Name="test-parameter-name", Type="String", Value="test-param-value"
            )
        )

        cluster_name = "cluster-name"
        service = "service-name"
        dependencies = [ssm_parameter]
        role = "role-arn"

        # Exercise
        stack = create_ecs_service_invalidation_stack(
            cluster=cluster_name,
            service=service,
            dependencies=dependencies,
            restart_role=role,
            timeout=30,
        )

        # Verify Lambda Function
        functions = [r for r in stack.Resources.values() if isinstance(r, Function)]
        assert len(functions) == 1, "There should be a Lambda Function resource"

        func = functions[0]
        assert re.search(
            r"restart.*ecs.*service", func.Properties.Handler, re.I
        ), "Lambda should restart an ECS service"
        assert func.Properties.Role == role

        # Verify Custom Resource
        custom_resources = [
            r for r in stack.Resources.values() if r["Type"].startswith("Custom::")
        ]
        assert (
            len(custom_resources) == 1
        ), "There should be a custom resource to restart the ECS service"

        restarter = custom_resources[0]
        assert restarter["Properties"]["ClusterArn"] == cluster_name
        assert restarter["Properties"]["ServiceArn"] == service

        # Verify dependencies for service restart
        dependent_values = _get_flattened_attributes(restarter)
        assert Ref(ssm_parameter) in dependent_values
        assert GetAtt(ssm_parameter, "Value") in dependent_values


def _get_flattened_attributes(value) -> set:
    """Get all attributes on this resource, flattening the hierarchy"""
    result = set()
    if isinstance(value, (set, list)):
        for v in value:
            result |= _get_flattened_attributes(v)
    elif isinstance(value, dict):
        for v in value.values():
            result |= _get_flattened_attributes(v)
    elif isinstance(value, AWSObject):
        for k in value:
            v = value[k]
            result |= _get_flattened_attributes(v)
    else:
        result.add(value)

    return result



"""Tools for converting configuration into SSM Parameters."""

import re
from typing import Any

import inflection
from flyingcircus.core import Stack
from flyingcircus.service.ssm import SSMParameter
from flyingcircus.service.ssm import SSMParameterProperties

#: RegEx to match `invalid characters
#: <https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/resources-section-structure.html#resources-section-structure-logicalid>`_
#: in a CloudFormation logical name
INVALID_LOGICAL_NAME_RE = re.compile(r"[^a-zA-Z0-9]+")

#: RegEx to match `invalid characters
#: <https://docs.aws.amazon.com/systems-manager/latest/userguide/sysman-parameter-name-constraints.html>`_
#: in one component of the path (aka name) for a SSM parameter.
#:
#: Note that a full path may contain a slash, but a single component may not.
INVALID_SSM_PARAMETER_COMPONENT_RE = re.compile(r"[^a-zA-Z0-9_.-]")


def convert_hierarchy_to_ssm(appconfig: dict) -> Stack:
    """Convert a hierarchical nested dictionary into SSM Parameters."""
    stack = Stack(Description="SSM Parameters")
    create_params_from_dict(stack, appconfig)
    return stack


def create_params_from_dict(
    stack: Stack, appconfig: dict, path_prefix: str = "/"
) -> None:
    for key, value in appconfig.items():
        _check_path_component_is_valid(key)
        item_path = path_prefix + key

        # Nested dictionaries form a parameter hierarchy
        if isinstance(value, dict):
            create_params_from_dict(stack, value, item_path + "/")
            continue

        # Store this value as a parameter
        logical_name = _clean_logical_name(item_path)
        logical_name = _dedupe_logical_name(stack, logical_name)

        if isinstance(value, list):
            # Store lists of plain values as a StringList
            stack.Resources[logical_name] = SSMParameter(
                Properties=SSMParameterProperties(
                    Name=item_path,
                    Type="StringList",
                    Value=_get_list_parameter_value(value),
                )
            )
        else:
            # Plain values should be stored as a string parameter
            stack.Resources[logical_name] = SSMParameter(
                Properties=SSMParameterProperties(
                    Name=item_path,
                    Type="String",
                    Value=_get_plain_parameter_value(value),
                )
            )


def _check_path_component_is_valid(component: str):
    """Check that this configuration key is valid to use as a component of a SSM Parameter Path.

    We don't want to alter provided configuration names, so we raise an error
    if they can't be used as-is.
    """
    if INVALID_SSM_PARAMETER_COMPONENT_RE.search(component):
        raise ValueError(f"Configuration has invalid key: {component}")


def _clean_logical_name(name: str) -> str:
    """Remove unsupported characters from a Cloud Formation logical name,
    and make it human-readable.
    """
    # We break the name into valid underscore-separated components, and then camelize it

    # Separate existing camelized words with underscore
    result = inflection.underscore(name)

    # Replace invalid characters with no more than 1 underscore
    result = INVALID_LOGICAL_NAME_RE.sub("_", result).strip("_")
    if not result:
        # There are no valid characters in the name, so we substitute in
        # some placeholder text that is valid
        result = "SymbolsOnly"

    # Turn into a CamelCase version using the underscores as word separators
    result = inflection.camelize(result)

    return result


def _dedupe_logical_name(stack: Stack, logical_name: str) -> str:
    """Create a unique logical name."""
    # We simply add a suffix to the existing name, if there is a conflict
    result = logical_name
    while result in stack.Resources:
        result += "Dupe"
    return result


def _get_list_parameter_value(value: list) -> str:
    """Lists of parameters should be stored as a comma-separated string."""
    if not value:
        raise ValueError("Cannot store an empty list in SSM Parameter Store")

    result = []
    for v in value:
        if isinstance(v, (set, list, dict)):
            raise ValueError(
                "Cannot store complex values inside a list in SSM Parameter Store"
            )

        cleaned = _get_plain_parameter_value(v)

        if not cleaned:
            raise ValueError(
                "Cannot store empty values inside a list in SSM Parameter Store"
            )
        if "," in cleaned:
            raise ValueError(
                "Cannot store values with a comma inside a list in SSM Parameter Store"
            )

        result.append(cleaned)

    return ",".join(result)


def _get_plain_parameter_value(value: Any) -> str:
    """All single-value parameters should be stored as a string.

    We normalise the outputs for some situations.
    """
    if isinstance(value, bool):
        return str(value).lower()
    if value is None:
        raise ValueError("Cannot store null values in SSM Parameter Store")
    return str(value)



"""Customised hypothesis stategies for generating test data."""

import hypothesis.strategies as st


@st.composite
def aws_logical_name_strategy(draw):
    """A strategy that produces a valid logical name for an AWS Stack object."""
    return draw(
        st.builds(
            lambda a, b: a + b,
            st.text("ABCDEFGHIJKLMNOPQRSTUVWXYZ", min_size=1, max_size=1),
            st.text("abcdefghijklmnnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"),
        )
    )


@st.composite
def parameter_name_strategy(draw):
    """A strategy that produces a valid parameter name component."""
    return draw(
        # NB: No "/", but otherwise it's the character set for a SSM Parameter Name
        st.from_regex(r"[a-zA-Z0-9_.-]+", fullmatch=True)
    )



import re

import hypothesis.strategies as st
import pytest
from flyingcircus.core import Stack
from flyingcircus.service.ssm import SSMParameter
from hypothesis import given

from ssmash.converter import convert_hierarchy_to_ssm
from .strategies import aws_logical_name_strategy
from .strategies import parameter_name_strategy

LOGICAL_NAME_RE = re.compile(r"[a-z0-9]+", re.I)
PARAMETER_NAME_RE = re.compile(r"[a-z0-9_./-]+", re.I)


class TestConvertHierarchyToSsm:
    """Tests for convert_hierarchy_to_ssm."""

    def _verify_stack_has_parameter(
        self, stack: Stack, path: str, value: str, logical_name: str = None
    ):
        # If we know the logical name, then just get the resource directly
        if logical_name:
            assert logical_name in stack.Resources
            assert stack.Resources[logical_name].Properties.Name == path
            assert stack.Resources[logical_name].Properties.Value == value
            return

        # Otherwise search for the Parameter by path
        params = {
            param.Properties.Name: param.Properties.Value
            for param in stack.Resources.values()
            if isinstance(param, SSMParameter)
        }

        assert path in params
        assert params[path] == value

    @given(aws_logical_name_strategy(), st.text(min_size=1))
    def test_should_create_parameter_resource_for_single_top_level_item(
        self, key, value
    ):
        # Exercise
        stack = convert_hierarchy_to_ssm({key: value})

        # Verify
        logical_name, param = stack.Resources.popitem()

        assert logical_name.lower() == key.lower()

        assert param.RESOURCE_TYPE == "AWS::SSM::Parameter"
        assert param.Properties.Name == "/" + key
        assert param.Properties.Type == "String"
        assert param.Properties.Value == value

    @pytest.mark.parametrize(
        ("value", "expected"), [(0, "0"), (True, "true"), (False, "false")]
    )
    def test_should_store_any_primitive_as_a_string(self, value, expected):
        # Exercise
        stack = convert_hierarchy_to_ssm({"some_key": value})

        # Verify
        _, param = stack.Resources.popitem()
        assert param.Properties.Type == "String"
        assert param.Properties.Value == expected

    @given(
        st.lists(
            st.one_of(
                st.text(alphabet=st.characters(blacklist_characters=","), min_size=1),
                st.integers(),
            ),
            min_size=1,
        )
    )
    def test_should_store_list_of_primitives_as_a_stringlist(self, value):
        # Exercise
        stack = convert_hierarchy_to_ssm({"some_key": value})

        # Verify
        _, param = stack.Resources.popitem()
        assert param.Properties.Type == "StringList"

        assert isinstance(param.Properties.Value, str)

        actual_list = param.Properties.Value.split(",")
        assert len(actual_list) == len(value)
        for expected, actual in zip(value, actual_list):
            assert str(expected) == actual

    def test_should_throw_error_for_none_value(self):
        # Exercise
        with pytest.raises(ValueError, match="(null|None)"):
            convert_hierarchy_to_ssm({"some_key": None})

    @pytest.mark.parametrize("name", ["some?value", "some/value"])
    def test_should_throw_error_for_invalid_parameter_name(self, name):
        # Exercise
        with pytest.raises(
            ValueError, match="invalid.*{}".format(name.replace("?", r"\?"))
        ):
            convert_hierarchy_to_ssm({name: "aaa"})

    @pytest.mark.parametrize(
        "value",
        [[], [None], [""], ["comma-ed,value"], [{"nested": "dict"}], [["nested_list"]]],
    )
    def test_should_throw_error_for_unsupported_list_contents(self, value):
        # Exercise
        with pytest.raises(ValueError, match="(list|null)"):
            convert_hierarchy_to_ssm({"some_key": value})

    @given(parameter_name_strategy())
    def test_should_clean_logical_names(self, key):
        # Exercise
        stack = convert_hierarchy_to_ssm({key: "some_value"})

        # Verify
        logical_name, param = stack.Resources.popitem()

        assert LOGICAL_NAME_RE.fullmatch(
            logical_name
        ), f"A logical name must be alphanumeric: '{logical_name}'"

        assert PARAMETER_NAME_RE.fullmatch(
            param.Properties.Name
        ), f"A parameter name must contain only valid characters: '{param.Properties.Name}'"
        assert (
            "/" not in param.Properties.Name[1:]
        ), "individual path components may not have a slash"

    def test_should_create_parameter_paths_from_nested_dictionaries(self):
        # Setup
        appconfig = {
            "top_value": "aaa",
            "top_dict": {"middle_value": "bbb", "middle_dict": {"bottom_value": "ccc"}},
        }

        # Exercise
        stack = convert_hierarchy_to_ssm(appconfig)

        # Verify
        assert (
            len(stack.Resources) == 3
        ), "3 parameter values should be stored from this input"

        self._verify_stack_has_parameter(stack, "/top_value", "aaa")
        self._verify_stack_has_parameter(stack, "/top_dict/middle_value", "bbb")
        self._verify_stack_has_parameter(
            stack, "/top_dict/middle_dict/bottom_value", "ccc"
        )

    def test_should_deterministically_handle_name_clashes_from_cleaned_logical_names(
        self
    ):
        # Setup
        appconfig = {
            "some": {"value": "eee"},
            "some-Value": "ccc",
            "some.Value": "ddd",
            "some_value": "aaa",
            "some_value_dupe": "fff",
            "someValue": "bbb",
        }

        # Exercise
        stack = convert_hierarchy_to_ssm(appconfig)

        # Verify
        assert (
            len(stack.Resources) == 6
        ), "6 parameter values should be stored from this input"

        self._verify_stack_has_parameter(stack, "/some/value", "eee", "SomeValue")
        self._verify_stack_has_parameter(stack, "/some-Value", "ccc", "SomeValueDupe")
        self._verify_stack_has_parameter(
            stack, "/some.Value", "ddd", "SomeValueDupeDupe"
        )
        self._verify_stack_has_parameter(
            stack, "/some_value", "aaa", "SomeValueDupeDupeDupe"
        )
        self._verify_stack_has_parameter(
            stack, "/some_value_dupe", "fff", "SomeValueDupeDupeDupeDupe"
        )
        self._verify_stack_has_parameter(
            stack, "/someValue", "bbb", "SomeValueDupeDupeDupeDupeDupe"
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("CHANGELOG.rst") as history_file:
    history = history_file.read()

requirements = ["click>=7.0,<8", "PyYAML==5.1", "flying-circus==0.6.5", "inflection==0.3.1"]

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest"]

setup(
    author="Gary Donovan",
    author_email="gazza@gazza.id.au",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development",
        "Topic :: System :: Systems Administration",
    ],
    description="SSM AppConfig Storage Helper",
    entry_points={"console_scripts": ["ssmash=ssmash.cli:create_stack"]},
    install_requires=requirements,
    license="GNU Affero General Public License v3",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="ssmash application-configuration configuration AWS cloudformation infrastructure-as-code",
    metadata_version="2.1",
    name="ssmash",
    package_dir={"": "src"},
    packages=find_packages(where="src", include=["ssmash"]),
    project_urls={
        "Documentation": "https://ssmash.readthedocs.io/en/latest/",
        "Source": "https://github.com/garyd203/ssmash",
        "Tracker": "https://github.com/garyd203/ssmash/issues",
    },
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/garyd203/ssmash",
    version="2.0.0",
    zip_safe=False,
)



"""Tools to invalidate applications that depend on the parameters."""

from typing import List

from flyingcircus.core import Stack
from flyingcircus.intrinsic_function import GetAtt, Ref
from flyingcircus.service.lambda_ import Function
from flyingcircus.service.ssm import SSMParameter

from ssmash.custom_resources import restart_ecs_service_resource_handler


def create_ecs_service_invalidation_stack(
    cluster,
    service,
    dependencies: List[SSMParameter],
    restart_role,
    timeout: int = 8 * 60,
) -> Stack:
    """Create CloudFormation resources to invalidate a single ECS service.

    This is accomplished by restarting the ECS service, which will force it to
    use the new parameters.

    Parameters:
        cluster: CloudFormation reference (eg. an ARN) to the cluster the service is in
        dependencies: SSM Parameters that this service uses
        service: CloudFormation reference (eg. an ARN) to the ECS service to invalidate
        restart_role: CloudFormation reference (eg. an ARN) to an IAM role
            that will be used to restart the ECS service.
        timeout: Number of seconds to wait for the ECS service to detect the
            new parameters successfully after restart. If we exceed this
            timeout it is presumed that the updated parameters are broken,
            and the changes will be rolled back. The default is 5 minutes,
            which should be enough for most services.
    """
    # TODO make restart_role optional, and create it on-the-fly if not provided
    # TODO find a way to share role and lambda between multiple calls in the same stack? can de-dupe/cache based on identity in the final stack
    # TODO get Lambda handler to have an internal timeout as well?

    stack = Stack(Description="Invalidate ECS service after parameter update")

    # Create an inline Lambda that can restart an ECS service, since this
    # isn't built-in CloudFormation functionality.
    stack.Resources[
        "RestartLambda"
    ] = restart_service_lambda = Function.create_from_python_function(
        handler=restart_ecs_service_resource_handler, Role=restart_role
    )

    # The Lambda timeout should be a bit longer than the restart timeout,
    # to give some leeway.
    restart_service_lambda.Properties.Timeout = timeout + 15

    # Create a custom resource to restart the ECS service.
    #
    # We don't want the service restart to happen until the parameters have
    # all been created, so we need to have the Restarter resource depend on
    # the parameters (either implicitly or via DependsOn). We also want the
    # restart to only happen if the parameters have actually changed - this
    # can be done if we make the SSM Parameters be part of the resource
    # specification (both the key and the value).
    stack.Resources["Restarter"] = dict(
        Type="Custom::RestartEcsService",
        Properties=dict(
            ServiceToken=GetAtt(restart_service_lambda, "Arn"),
            ClusterArn=cluster,
            ServiceArn=service,
            IgnoredParameterNames=[Ref(p) for p in dependencies],
            IgnoredParameterKeys=[GetAtt(p, "Value") for p in dependencies],
        ),
    )

    # TODO consider creating a waiter anyway, so that the timeout is strictly reliable

    return stack
